from typing import Dict


class L10NRepository(object):

    def __init__(
        self,
        l10n_params: Dict[str, str],
    ) -> None:

        self.l10n_params = l10n_params

    def get_currency_symbol(
        self,
    ) -> str:

        return self.l10n_params["L10N_CURRENCY_SYMBOL"]

    def get_currency_symbol_position(
        self,
    ) -> str:

        return self.l10n_params["L10N_CURRENCY_SYMBOL_POSITION"]

    def get_decimal_places(
        self,
    ) -> str:

        return self.l10n_params["L10N_DECIMAL_PLACES"]

    def get_decimal_separator(
        self,
    ) -> str:

        return self.l10n_params["L10N_DECIMALS_SEPARATOR"]

    def get_thousands_separator(
        self,
    ) -> str:

        return self.l10n_params["L10N_THOUSANDS_SEPARATOR"]

    def get_datetime_format(
        self,
    ) -> str:

        return self.l10n_params["L10_PYTHON_DATETIME_FORMAT"]
