from sysactions import get_cfg

ask_pager_number = {}


def should_ask_pager_number(
    pos_id,  # type: str
):
    # type: (...) -> bool

    global ask_pager_number

    if pos_id not in ask_pager_number:
        ask_pager_number[str(pos_id)] = get_cfg(pos_id).find_value("AskPagerNumber", "false").lower() == "true"

    return ask_pager_number[pos_id]
