class PosSaleTypeNavigation:
    def __init__(
        self,
        pos_id,             # type: int
        sale_type,          # type: str
        navigation_id,      # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.sale_type = sale_type
        self.navigation_id = navigation_id
