from messagebus import (
    TokenCreator,
    TokenPriority,
)

TK_I18N_GETMSG = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="7",
    number="2",
)
TK_I18N_GETLANGS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="7",
    number="4"
)
TK_I18N_GETTABLEJSON = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="7",
    number="5",
)
