from application.model.configuration import Configurations


class LoyaltyIdentifierInteractor(object):
    def __init__(
        self,
        configs,    # type: Configurations
    ):
        # type: (...) -> None

        self.logger = configs.logger
        self._cache = {
            'LoyaltyIdentifierType': configs.id_type,
        }

    def get_key(
        self,
        key,    # type: str
    ):
        # type: (...) -> str

        return self._cache.get(key, None)
