class OrderPicture(object):
    def __init__(
        self,
        store_id,      # type: str
        order_info,    # type: dict
    ):
        self.store_code = store_id
        self.order_info = order_info
