from typing import Optional


class PartCode(object):
    def __init__(
        self,
        id,                     # type: str
        context,                # type: str
        quantity_sold=None,     # type: Optional[int]
    ):
        # type: (...) -> None

        self.id = id
        self.context = context
        self.quantity_sold = quantity_sold
