class ErrorGettingPOSModel(Exception):

    def __init__(
        self,
        pos_id: int,
    ) -> None:
        super(ErrorGettingPOSModel, self).__init__(f"Could not retrieve state of pos id {pos_id}")
